#include "fire.h"
#include <string.h>

#define MaxAge 256
#define MaxDepth 256

#define AgeDepthMix(x,y) \
	(((Age - (x)) & (MaxAge - 1)) * MaxDepth   + (MaxDepth - ((y) + 1)))

void IncrementAge(void)
{
    Age += 1;

    if (Age == MaxAge)
        Age = 0;
}

static uint64 HashSize = 0x400000;
static boolean FlagHashInit = 0;
void HashClear(void)
{
    int i;
    memset(HashTable, 0, HashSize * sizeof(typeHash));
    memset(PVHashTable, 0, 0x10000 * sizeof(typePVHash));
    for (i = 0; i < HashSize; i++)
        (HashTable + i)->age = (MaxAge / 2);
    Age = 0;
}
int InitHash(int mb)
{
    Age = 0;
    HashSize = ((1ULL << MSB(mb)) << 20) / sizeof(typeHash);

    if (HashSize > 0x100000000)
        HashSize = 0x100000000;
    mb = (HashSize * sizeof(typeHash)) >> 20;
    HashMask = HashSize - 4;

    if (FlagHashInit)
        AlignedFree(HashTable);
    FlagHashInit = true;
    MemAlign(HashTable, 64, HashSize * sizeof(typeHash));
    HashClear();
    return mb;
}
void HashLowerALL(typePos *Position, int move, int depth, int Value)
{
    int Depth, i, k = Position->Dyn->Hash & HashMask;
    typeHash *trans;
    int max = 0, w = 0;
    move &= 0x7fff;

    for (i = 0; i < 4; i++)
    {
        trans = HashTable + (k + i);

        if ((trans->hash ^ (Position->Dyn->Hash >> 32)) == 0 && (!trans->DepthLower || IsAll(trans))
                && trans->DepthLower <= depth)
        {
            trans->DepthLower = depth;
            trans->move = move;
            trans->ValueUpper = Value;
            trans->age = Age;
            trans->flags |= FlagLower | FlagAll;
            return;
        }
        Depth = MAX(trans->DepthLower, trans->DepthUpper);

        if (AgeDepthMix(trans->age, Depth) > max)
        {
            max = AgeDepthMix(trans->age, Depth);
            w = i;
        }
    }
    trans = HashTable + (k + w);
    trans->hash = (Position->Dyn->Hash >> 32);
    trans->DepthUpper = 0;
    trans->ValueLower = 0;
    trans->DepthLower = depth;
    trans->move = move;
    trans->ValueUpper = Value;
    trans->age = Age;
    trans->flags = FlagLower | FlagAll;
    return;
}
void HashUpperCUT(typePos *Position, int depth, int Value)
{
    int Depth, i, k = Position->Dyn->Hash & HashMask;
    typeHash *trans;
    int max = 0, w = 0;

    for (i = 0; i < 4; i++)
    {
        trans = HashTable + (k + i);

        if (!(trans->hash ^ (Position->Dyn->Hash >> 32)) && (!trans->DepthUpper || IsCut(trans))
                && trans->DepthUpper <= depth)
        {
            trans->DepthUpper = depth;
            trans->ValueLower = Value;
            trans->age = Age;
            trans->flags |= FlagUpper | FlagCut;
            return;
        }
        Depth = MAX(trans->DepthLower, trans->DepthUpper);

        if (AgeDepthMix(trans->age, Depth) > max)
        {
            max = AgeDepthMix(trans->age, Depth);
            w = i;
        }
    }
    trans = HashTable + (k + w);
    trans->hash = (Position->Dyn->Hash >> 32);
    trans->DepthLower = 0;
    trans->move = 0;
    trans->ValueUpper = 0;
    trans->DepthUpper = depth;
    trans->ValueLower = Value;
    trans->age = Age;
    trans->flags = FlagUpper | FlagCut;
    return;
}
void HashLower(uint64 Z, int move, int depth, int Value)
{
    int Depth, i, k = Z & HashMask;
    typeHash *trans;
    int max = 0, w = 0;
    move &= 0x7fff;

    for (i = 0; i < 4; i++)
    {
        trans = HashTable + (k + i);

        if (!(trans->hash ^ (Z >> 32)) && !IsExact(trans) && trans->DepthLower <= depth)
        {
            trans->DepthLower = depth;
            trans->move = move;
            trans->ValueUpper = Value;
            trans->age = Age;
            trans->flags |= FlagLower;
            trans->flags &= ~FlagAll;
            return;
        }
        Depth = MAX(trans->DepthLower, trans->DepthUpper);

        if (AgeDepthMix(trans->age, Depth) > max)
        {
            max = AgeDepthMix(trans->age, Depth);
            w = i;
        }
    }
    trans = HashTable + (k + w);
    trans->hash = (Z >> 32);
    trans->DepthUpper = 0;
    trans->ValueLower = 0;
    trans->DepthLower = depth;
    trans->move = move;
    trans->ValueUpper = Value;
    trans->age = Age;
    trans->flags = FlagLower;
    return;
}
void HashUpper(uint64 Z, int depth, int Value)
{
    int Depth, i, k = Z & HashMask;
    typeHash *trans;
    int max = 0, w = 0;

    for (i = 0; i < 4; i++)
    {
        trans = HashTable + (k + i);

        if (!(trans->hash ^ (Z >> 32)) && !IsExact(trans) && trans->DepthUpper <= depth)
        {
            trans->DepthUpper = depth;
            trans->ValueLower = Value;
            trans->age = Age;
            trans->flags |= FlagUpper;
            trans->flags &= ~FlagCut;
            return;
        }
        Depth = MAX(trans->DepthLower, trans->DepthUpper);

        if (AgeDepthMix(trans->age, Depth) > max)
        {
            max = AgeDepthMix(trans->age, Depth);
            w = i;
        }
    }
    trans = HashTable + (k + w);
    trans->hash = (Z >> 32);
    trans->DepthLower = 0;
    trans->move = 0;
    trans->ValueUpper = 0;
    trans->DepthUpper = depth;
    trans->ValueLower = Value;
    trans->age = Age;
    trans->flags = FlagUpper;
    return;
}
static void PVHash(typePos *Position, int move, int depth, int Value)
{
    int i, k = Position->Dyn->Hash & PVHashMask;
    typePVHash *trans;
    int w = 0, max = 0;

    for (i = 0; i < 4; i++)
    {
        trans = PVHashTable + (k + i);

        if (trans->hash == Position->Dyn->Hash)
        {
            trans->depth = depth;
            trans->Value = Value;
            trans->move = move;
            trans->age = Age;
            return;
        }

        if (AgeDepthMix(trans->age, trans->depth) > max)
        {
            max = AgeDepthMix(trans->age, trans->depth);
            w = i;
        }
    }
    trans = PVHashTable + (k + w);
    trans->hash = Position->Dyn->Hash;
    trans->depth = depth;
    trans->move = move;
    trans->Value = Value;
    trans->age = Age;
}

void HashExact(typePos *Position, int move, int depth, int Value, int Flags)
{
    int Depth, i, j, k = Position->Dyn->Hash & HashMask;
    typeHash *trans;
    int max = 0, w = 0;
    move &= 0x7fff;
    PVHash(Position, move, depth, Value);

    for (i = 0; i < 4; i++)
    {
        trans = HashTable + (k + i);

        if ((trans->hash ^ (Position->Dyn->Hash >> 32)) == 0 && MAX(trans->DepthUpper, trans->DepthLower) <= depth)
        {
            trans->DepthUpper = trans->DepthLower = depth;
            trans->move = move;
            trans->ValueLower = trans->ValueUpper = Value;
            trans->age = Age;
            trans->flags = Flags;

            for (j = i + 1; j < 4; j++)
            {
                trans = HashTable + (k + j);

                if ((trans->hash ^ (Position->Dyn->Hash >> 32)) == 0
                        && MAX(trans->DepthUpper, trans->DepthLower) <= depth)
                {
                    memset(trans, 0, 16);
                    trans->age = Age ^ (MaxAge / 2);
                }
            }
            return;
        }
        Depth = MAX(trans->DepthLower, trans->DepthUpper);

        if (AgeDepthMix(trans->age, Depth) > max)
        {
            max = AgeDepthMix(trans->age, Depth);
            w = i;
        }
    }
    trans = HashTable + (k + w);
    trans->hash = (Position->Dyn->Hash >> 32);
    trans->DepthUpper = trans->DepthLower = depth;
    trans->move = move;
    trans->ValueLower = trans->ValueUpper = Value;
    trans->age = Age;
    trans->flags = Flags;
    return;
}
